@@ -3,7 +3,9 @@
 #define _GNU_SOURCE
 #include <errno.h>
 #include <fcntl.h>
+#include <grp.h>
 #include <libgen.h>
+#include <pwd.h>
 #include <sched.h>
 #include <semaphore.h>
 #include <stdio.h>
@@ -47,13 +49,16 @@ const char *VERBOSE_LEVELS[] = { "error", "warning", "info", "debug" };
 
 /* Default bind-mounts. */
 const char *DEFAULT_BINDS[] = { "/dev",
-                                "/etc/passwd",
-                                "/etc/group",
                                 "/etc/hosts",
                                 "/etc/resolv.conf",
                                 "/proc",
                                 "/sys",
                                 NULL };
+struct bind BINDS_OPTIONAL[] = {
+   { "/var/opt/cray/alps/spool", "/var/opt/cray/alps/spool" },
+   { "/var/lib/hugetlbfs",       "/var/opt/cray/hugetlbfs" },
+   { NULL, NULL }
+};
 
 
 /** External variables **/
@@ -87,12 +92,29 @@ void join_end();
 void log_ids(const char *func, int line);
 void sem_timedwait_relative(sem_t *sem, int timeout);
 void setup_namespaces(struct container *c);
+void setup_passwd(struct container *c);
 
 
 /** Functions **/
 
+/* Bind-mount a null-terminated array of struct bind objects. */
+void bind_mounts(struct bind *binds, char * newroot,
+                 enum bind_dep dep, unsigned long flags)
+{
+   char *dst;
+
+   for (int i = 0; binds[i].src != NULL; i++) {
+      T_ (1 <= asprintf(&dst, "%s%s", newroot, binds[i].dst));
+      if (dep == BD_OPTIONAL && !(path_exists(binds[i].src) && path_exists(dst)))
+         continue;
+      Zf (mount(binds[i].src, dst, NULL, MS_REC|MS_BIND|flags, NULL),
+          "can't bind %s to %s", binds[i].src, dst);
+   }
+}
+
+
 /* Set up new namespaces or join existing namespaces. */
-void containerize(struct container * c)
+void containerize(struct container *c)
 {
    if (c->join)
       join_begin(c->join_ct, c->join_tag);
@@ -131,6 +153,10 @@ void enter_udss(struct container *c)
       Zf (mount(DEFAULT_BINDS[i], path, NULL, MS_REC|MS_BIND|MS_RDONLY, NULL),
           "can't bind %s to %s", (char *) DEFAULT_BINDS[i], path);
    }
+   // Bind-mount optional files and directories if host and guest paths exist.
+   bind_mounts(BINDS_OPTIONAL, c->newroot, BD_OPTIONAL, MS_RDONLY);
+   // /etc/passwd and /etc/group
+   setup_passwd(c);
    // Container /tmp
    T_ (1 <= asprintf(&path, "%s%s", c->newroot, "/tmp"));
    if (c->private_tmp) {
@@ -353,6 +379,18 @@ void msg(int level, char *file, int line, int errno_, char *fmt, ...)
       exit(EXIT_FAILURE);
 }
 
+/* Return true if the given path exists, false otherwise. On error, exit. */
+bool path_exists(char * path) {
+   struct stat sb;
+
+   if (!stat(path, &sb))
+      return true;
+   else if (errno != ENOENT)
+      Tf (0, "can't stat: %s", path);
+
+   return false;
+}
+
 /* Replace the current process with user command and arguments. */
 void run_user_command(char *argv[], char *initial_dir)
 {
@@ -427,6 +465,53 @@ void setup_namespaces(struct container *c)
    LOG_IDS;
 }
 
+/* Build /etc/passwd and /etc/group files and bind-mount them into newroot. We
+   do it this way so that we capture the relevant host username and group name
+   mappings regardless of where they come from. (We used to simply bind-mount
+   the host's /etc/passwd and /etc/group, but this fails for LDAP at least;
+   see issue #212.) After bind-mounting, we remove them on the host side;
+   they'll persist inside the container and then disappear completely when the
+   latter exits. */
+void setup_passwd(struct container *c)
+{
+   int fd;
+   char *path, *newpath;
+   struct group *g;
+   struct passwd *p;
+
+   // /etc/passwd
+   T_ (path = strdup("/tmp/ch-run_passwd.XXXXXX"));
+   T_ (-1 != (fd = mkstemp(path)));
+   if (c->container_uid != 0)
+      T_ (1 <= dprintf(fd, "root:x:0:0:root:/root:/bin/sh\n"));
+   if (c->container_uid != 65534)
+      T_ (1 <= dprintf(fd, "nobody:x:65534:65534:nobody:/:/bin/false\n"));
+   T_ (p = getpwuid(c->container_uid));
+   T_ (1 <= dprintf(fd, "%s:x:%u:%u:%s:/home/%s:/bin/sh\n",
+                    p->pw_name, c->container_uid, c->container_gid,
+                    p->pw_gecos, getenv("USER")));
+   Z_ (close(fd));
+   T_ (1 <= asprintf(&newpath, "%s/etc/passwd", c->newroot));
+   Zf (mount(path, newpath, NULL, MS_BIND, NULL),
+       "can't bind %s to %s", path, newpath);
+   Z_ (unlink(path));
+
+   // /etc/group
+   T_ (path = strdup("/tmp/ch-run_group.XXXXXX"));
+   T_ (-1 != (fd = mkstemp(path)));
+   if (c->container_gid != 0)
+      T_ (1 <= dprintf(fd, "root:x:0:\n"));
+   if (c->container_gid != 65534)
+      T_ (1 <= dprintf(fd, "nogroup:x:65534:\n"));
+   T_ (g = getgrgid(c->container_gid));
+   T_ (1 <= dprintf(fd, "%s:x:%u:\n", g->gr_name, c->container_gid));
+   Z_ (close(fd));
+   T_ (1 <= asprintf(&newpath, "%s/etc/group", c->newroot));
+   Zf (mount(path, newpath, NULL, MS_BIND, NULL),
+       "can't bind %s to %s", path, newpath);
+   Z_ (unlink(path));
+}
+
 /* Report the version number. */
 void version(void)
 {
